// ===============================================================
//                 ORBITER MODULE: PalmRControl
//               Copyright(C) 2006 Rodolfo Turco
//       It's released under the GNU General Public License
//              http://www.webalice.it/rodolfo.turco
//               rodolfo.turco[ a T ] virgilio [dot] it
//                          ver. 1.0b1
//
// PalmRControl.cpp
//
// This module is the Windows side part of the Palm Remote Control
//
// ===============================================================

//TODO ver. 1.0
// * May crash on setfocus (always on GL-02; on GL-01 if was on GL-01 with focus on SH-01 and close/open)
//   opcFocusChanged may conflit with oapiSetFocusObject, check it
// * use oapiIsVessel to check the handle
// * sometime it enters in a state that do not send commands to the palm

//TODO ver. 1.1
// * Let SUSTAINMAXCOUNT settable
// * Per buffer receiving, now is single character
// * Control/strumentation forms

#define STRICT
#define ORBITER_MODULE
#include "windows.h"
#include "orbitersdk.h"
#include "resource.h"
#include <stdio.h>
#include <stdlib.h>
#include <process.h>

// ==============================================================
// Global variables and defines
// ==============================================================
#define COMMANDTERMINATOR   (unsigned char)23 // it's the command terminator (ASCII CODE 23)
#define SUSTAINMAXCOUNT     3 // for a uniform RCS control a command is repeated SUSTAINMAXCOUNT steps
#define RXBUFFERSIZE        16
#define COMMANDBUFFERSIZE   1024

struct GPARAMS
{
  HANDLE hComm;
#ifdef _DEBUG
  HANDLE hFile;
#endif

  HINSTANCE g_hInst;  // module instance handle
  DWORD     g_dwCmd;  // custom function identifier
  VESSEL    *pVessel;
  OBJHANDLE hVessel;

  COMMTIMEOUTS CommDefaultTimeouts;
  COMMTIMEOUTS CommTimeouts;

  bool   bCOMOpened;
  bool   bExitCOMThread;
  bool   bSustainOn;
  short  nSustainCount;
  short  nSustain;
  short  nRCSMode;
  int    iComNum;
  double rcslevel;
  double prevSimT;
  double lThGroup_MAIN;
  double lThGroup_RETRO;
  double lThGroup_HOVER;
};

static GPARAMS g_Param;

// ==============================================================
// COM commands
// ==============================================================
// The command data format is an ASCII string in "cpt" format where:
        // c: are a two chars command e.g. "Ho" for Hover
        // p: is the parameter (if required) e.g. "12" for Hover value in 0-100 range
        // t: is the command terminator (not the question mark, it's the ASCII CODE 23)
        // e.g. Ho12? the set the Hover to 12/100

// To Palm (inverted)
#define COMMAND_SV 0x5356 // Set Vessel list

// From/To Palm (inverted)
#define COMMAND_CO 0x4F43 // Set COM Open (advide the other side it's ready to communicate)
#define COMMAND_CL 0x4C43 // Set COM Close (advide the other side it closed the connection)
#define COMMAND_Ho 0x6F48 // +### Set Hover
#define COMMAND_Ma 0x614D // +### Set Main
#define COMMAND_Rt 0x7452 // +### Set Retro
#define COMMAND_KR 0x524B // Set KillRot
#define COMMAND_HL 0x4C48 // Set HLevel
#define COMMAND_HA 0x4148 // Set HAlt
#define COMMAND_Pr 0x7250 // Set Prograde
#define COMMAND_Re 0x6552 // Set Retrograde
#define COMMAND_NP 0x504E // Set Normal +
#define COMMAND_NM 0x4D4E // Set Normal -
#define COMMAND_Le 0x654C // +### Set RCS Level
#define COMMAND_SR 0x5253 // Set RCS Mode (0=disable, 1=rotational, 2=linear)
#define COMMAND_Se 0x6553 // Select vessel

// From Palm (inverted)
#define COMMAND_RD 0x4452 // Set RCS Rot D
#define COMMAND_RU 0x5552 // Set RCS Rot U
#define COMMAND_Rl 0x6C52 // Set RCS Rot Left ^
#define COMMAND_Rr 0x7252 // Set RCS Rot Right ^
#define COMMAND_RL 0x4C52 // Set RCS Rot Left
#define COMMAND_RR 0x5252 // Set RCS Rot Right
#define COMMAND_RX 0x5852 // Set RCS Rot X
#define COMMAND_LD 0x444C // Set RCS Lin D
#define COMMAND_LU 0x554C // Set RCS Lin U
#define COMMAND_LF 0x464C // Set RCS Lin F
#define COMMAND_LB 0x424C // Set RCS Lin B
#define COMMAND_LL 0x4C4C // Set RCS Lin L
#define COMMAND_LR 0x524C // Set RCS Lin R
#define COMMAND_SF 0x4653 // Set Focus to vessel
#define COMMAND_GV 0x5647 // Get Vessel list
#define COMMAND_GH 0x4847 // Get Hover
#define COMMAND_GM 0x4D47 // Get Main
#define COMMAND_GR 0x5247 // Get Retro
#define COMMAND_GL 0x4C47 // Get RCS Level
#define COMMAND_Gi 0x6947 // Get KillRot
#define COMMAND_Ge 0x6547 // Get HLevel
#define COMMAND_Gh 0x6847 // Get HAlt
#define COMMAND_Go 0x6F47 // Get Prograde
#define COMMAND_Gt 0x7447 // Get Retrograde
#define COMMAND_Gp 0x7047 // Get Normal +
#define COMMAND_Gm 0x6D47 // Get Normal -
#define COMMAND_GT 0x5447 // Get RCS type
#define COMMAND_GF 0x4647 // Get Focused vessel

// ==============================================================
// Local prototypes
// ==============================================================
void OpenDlgClbk(void *context);
BOOL CALLBACK MsgProc(HWND, UINT, WPARAM, LPARAM);
void COMSendSimplex(short key);
void COMSendNumber(short key, long lNumber);
void COMSendString(short key, char* pStringa);
void sendNewNavmode(int mode, bool active);
void setNewNavmode(int mode, bool active);

// ==============================================================
// Serial data receive thread
// Incoming commands are executed here
// ==============================================================
void COMThread( void *pArg )
{
#ifdef _DEBUG
DWORD nNumberOfBytesWritten;
#endif
  DWORD nNumberOfBytesToRead;
  DWORD nNumberOfBytesRead;
  long  lTotReadByte;
  long  lParam;
  long  lIx, lNum;
  unsigned short usCommand;
  char* pName;
  char  cName[128];
  OBJHANDLE hVessel2;
  unsigned char cStringa[4096];
  unsigned char cRxBuffer[RXBUFFERSIZE];
  unsigned char cCommandBuffer[COMMANDBUFFERSIZE];

  try
  {
    GPARAMS *pg_Param = (GPARAMS *)pArg;

    if(pg_Param == NULL)
      return;

  cRxBuffer[0] = 0;
  cCommandBuffer[0] = 0;
  nNumberOfBytesToRead = 1;
  lTotReadByte = 0;

#ifdef _DEBUG
pg_Param->hFile = CreateFile("PalmRCtrlLog.txt",GENERIC_WRITE,0,0,CREATE_ALWAYS,0,0);
#endif

  COMSendSimplex(COMMAND_CO); // Ready to COM

  while(true)
  {
    // wait rx data with timeout
    ReadFile(pg_Param->hComm, cRxBuffer, nNumberOfBytesToRead, &nNumberOfBytesRead, NULL);

    // check exit condition
    if(pg_Param->bExitCOMThread)
    {
      COMSendSimplex(COMMAND_CL);
      Sleep(200);
//      PurgeComm(pg_Param->hComm, PURGE_RXCLEAR);
//      PurgeComm(pg_Param->hComm, PURGE_TXCLEAR);
      SetCommTimeouts(pg_Param->hComm, &pg_Param->CommDefaultTimeouts);// revert to default
      CloseHandle(pg_Param->hComm);// close serial COM
      pg_Param->hComm = INVALID_HANDLE_VALUE;
      break;//exit while(true)
    }
    else if(nNumberOfBytesRead)// we got data
    {
#ifdef _DEBUG
if(pg_Param->hFile)
  WriteFile(pg_Param->hFile,cRxBuffer,nNumberOfBytesRead,&nNumberOfBytesWritten,NULL);
#endif

      if(cRxBuffer[0] == COMMANDTERMINATOR)// if command terminator found then exec cCommandBuffer
      {
        if((lTotReadByte >= 2) && pg_Param->pVessel)
        {
          usCommand = *(unsigned short *)cCommandBuffer; // interpreted as short for speed

/*#ifdef _DEBUG
if(hFile)
{
  cRxBuffer[0] = '{';
  nNumberOfBytesRead = 1;
  WriteFile(hFile,cRxBuffer,nNumberOfBytesRead,&nNumberOfBytesWritten,NULL);
  WriteFile(hFile,cCommandBuffer,lTotReadByte,&nNumberOfBytesWritten,NULL);
  cRxBuffer[0] = '}';
  nNumberOfBytesRead = 1;
  WriteFile(hFile,cRxBuffer,nNumberOfBytesRead,&nNumberOfBytesWritten,NULL);
}
#endif*/
          switch(usCommand)
          {
          case COMMAND_Ho: // Hover
            if(lTotReadByte > 2)
            {
              lParam = atol((const char *)&(cCommandBuffer[2]));
              pg_Param->pVessel->SetThrusterGroupLevel(THGROUP_HOVER, lParam * 0.01);
              pg_Param->lThGroup_HOVER = lParam;
            }
            break;
          case COMMAND_Ma: // Main
            if(lTotReadByte > 2)
            {
              lParam = atol((const char *)&(cCommandBuffer[2]));
              pg_Param->pVessel->SetThrusterGroupLevel(THGROUP_MAIN, lParam * 0.01);
              pg_Param->lThGroup_MAIN = lParam;
            }
            break;
          case COMMAND_Rt: // Retro
            if(lTotReadByte > 2)
            {
              lParam = atol((const char *)&(cCommandBuffer[2]));
              pg_Param->pVessel->SetThrusterGroupLevel(THGROUP_RETRO, lParam * 0.01);
              pg_Param->lThGroup_RETRO = lParam;
            }
            break;
          case COMMAND_SF:
            if(pg_Param->hVessel)
              oapiSetFocusObject(pg_Param->hVessel);
            break;
          case COMMAND_Se: // Set vessel
            pName = (char *)&(cCommandBuffer[2]);
            pg_Param->hVessel = oapiGetVesselByName(pName);
            if(pg_Param->hVessel)
            {
				      pg_Param->pVessel = oapiGetVesselInterface(pg_Param->hVessel);
            }
            else
            {
              pg_Param->pVessel = oapiGetFocusInterface();
              pg_Param->hVessel = pg_Param->pVessel->GetHandle();
            }
            break;
          case COMMAND_KR: // KillRot
            setNewNavmode(NAVMODE_KILLROT, (atol((const char *)&(cCommandBuffer[2])))?true:false);
            if(atol((const char *)&(cCommandBuffer[2])) == 1)
            {
              pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_ELEVATOR, 0);
              pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_AILERON, 0);
              pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_RUDDER, 0);
            }
            break;
          case COMMAND_HL: // Set HLevel
            setNewNavmode(NAVMODE_HLEVEL, (atol((const char *)&(cCommandBuffer[2])))?true:false);
            break;
          case COMMAND_HA: // Set HAlt
            setNewNavmode(NAVMODE_HOLDALT, (atol((const char *)&(cCommandBuffer[2])))?true:false);
            break;
          case COMMAND_Pr: // Set Prograde
            setNewNavmode(NAVMODE_PROGRADE, (atol((const char *)&(cCommandBuffer[2])))?true:false);
            break;
          case COMMAND_Re: // Set Retrograde
            setNewNavmode(NAVMODE_RETROGRADE, (atol((const char *)&(cCommandBuffer[2])))?true:false);
            break;
          case COMMAND_NP: // Set Normal +
            setNewNavmode(NAVMODE_NORMAL, (atol((const char *)&(cCommandBuffer[2])))?true:false);
            break;
          case COMMAND_NM: // Set Normal -
            setNewNavmode(NAVMODE_ANTINORMAL, (atol((const char *)&(cCommandBuffer[2])))?true:false);
            break;
          case COMMAND_RD: // Set RCS Rot D
            pg_Param->pVessel->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_PITCHDOWN, pg_Param->rcslevel);
            pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_ELEVATOR, -pg_Param->rcslevel);
            pg_Param->bSustainOn = true;
            pg_Param->nSustainCount = 0;
            pg_Param->nSustain = THGROUP_ATT_PITCHDOWN;
            break;
          case COMMAND_RU: // Set RCS Rot U
            pg_Param->pVessel->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_PITCHUP, pg_Param->rcslevel);
            pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_ELEVATOR, pg_Param->rcslevel);
            pg_Param->bSustainOn = true;
            pg_Param->nSustainCount = 0;
            pg_Param->nSustain = THGROUP_ATT_PITCHUP;
            break;
          case COMMAND_Rl: // Set RCS Rot Left ^
            pg_Param->pVessel->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_BANKLEFT, pg_Param->rcslevel);
            pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_AILERON, -pg_Param->rcslevel);
            pg_Param->bSustainOn = true;
            pg_Param->nSustainCount = 0;
            pg_Param->nSustain = THGROUP_ATT_BANKLEFT;
            break;
          case COMMAND_Rr: // Set RCS Rot Right ^
            pg_Param->pVessel->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_BANKRIGHT, pg_Param->rcslevel);
            pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_AILERON, pg_Param->rcslevel);
            pg_Param->bSustainOn = true;
            pg_Param->nSustainCount = 0;
            pg_Param->nSustain = THGROUP_ATT_BANKRIGHT;
            break;
          case COMMAND_RL: // Set RCS Rot Left
            pg_Param->pVessel->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_YAWLEFT, pg_Param->rcslevel);
            pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_RUDDER, -pg_Param->rcslevel);
            pg_Param->bSustainOn = true;
            pg_Param->nSustainCount = 0;
            pg_Param->nSustain = THGROUP_ATT_YAWLEFT;
            break;
          case COMMAND_RR: // Set RCS Rot Right
            pg_Param->pVessel->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_YAWRIGHT, pg_Param->rcslevel);
            pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_RUDDER, pg_Param->rcslevel);
            pg_Param->bSustainOn = true;
            pg_Param->nSustainCount = 0;
            pg_Param->nSustain = THGROUP_ATT_YAWRIGHT;
            break;
          case COMMAND_RX: // Set RCS Rot X
            if(pg_Param->nRCSMode == NAVMODE_KILLROT)
            {
              setNewNavmode(NAVMODE_KILLROT, false);
            }
            else
            {
              setNewNavmode(NAVMODE_KILLROT, true);
              pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_ELEVATOR, 0);
              pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_AILERON, 0);
              pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_RUDDER, 0);
            }
            break;
          case COMMAND_LD: // Set RCS Lin D
            pg_Param->pVessel->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_DOWN, pg_Param->rcslevel);
            pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_ELEVATOR, -pg_Param->rcslevel);
            pg_Param->bSustainOn = true;
            pg_Param->nSustainCount = 0;
            pg_Param->nSustain = THGROUP_ATT_DOWN;
            break;
          case COMMAND_LU: // Set RCS Lin U
            pg_Param->pVessel->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_UP, pg_Param->rcslevel);
            pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_ELEVATOR, pg_Param->rcslevel);
            pg_Param->bSustainOn = true;
            pg_Param->nSustainCount = 0;
            pg_Param->nSustain = THGROUP_ATT_UP;
            break;
          case COMMAND_LF: // Set RCS Lin F
            pg_Param->pVessel->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_FORWARD, pg_Param->rcslevel);
            pg_Param->bSustainOn = true;
            pg_Param->nSustainCount = 0;
            pg_Param->nSustain = THGROUP_ATT_FORWARD;
            break;
          case COMMAND_LB: // Set RCS Lin B
            pg_Param->pVessel->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_BACK, pg_Param->rcslevel);
            pg_Param->bSustainOn = true;
            pg_Param->nSustainCount = 0;
            pg_Param->nSustain = THGROUP_ATT_BACK;
            break;
          case COMMAND_LL: // Set RCS Lin L
            pg_Param->pVessel->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_LEFT, pg_Param->rcslevel);
            pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_RUDDER, -pg_Param->rcslevel);
            pg_Param->bSustainOn = true;
            pg_Param->nSustainCount = 0;
            pg_Param->nSustain = THGROUP_ATT_LEFT;
            break;
          case COMMAND_LR: // Set RCS Lin R
            pg_Param->pVessel->IncThrusterGroupLevel_SingleStep(THGROUP_ATT_RIGHT, pg_Param->rcslevel);
            pg_Param->pVessel->SetControlSurfaceLevel(AIRCTRL_RUDDER, pg_Param->rcslevel);
            pg_Param->bSustainOn = true;
            pg_Param->nSustainCount = 0;
            pg_Param->nSustain = THGROUP_ATT_RIGHT;
            break;
          case COMMAND_Le: // Set RCS Level
            if(lTotReadByte > 2)
            {
              pg_Param->rcslevel = atol((const char *)&(cCommandBuffer[2]))/100.0;
            }
            break;
          case COMMAND_SR: // Set RCS Mode (0=disable, 1=rotational, 2=linear)
            switch(atol((const char *)&(cCommandBuffer[2])))
            {
              case 0:
              oapiSetAttitudeMode(pg_Param->hVessel, RCS_NONE);
              break;
              case 1:
              oapiSetAttitudeMode(pg_Param->hVessel, RCS_ROT);
              break;
              case 2:
              oapiSetAttitudeMode(pg_Param->hVessel, RCS_LIN);
              break;
            }
            break;
          case COMMAND_GV: // Get Vessel list
	          lNum = oapiGetVesselCount();
            cStringa[0] = 0;
	          for(lIx = 0; lIx < lNum; lIx++)
            {
		          hVessel2 = oapiGetVesselByIndex(lIx);
		          oapiGetObjectName(hVessel2, cName, 128);
              strcat_s((char *)cStringa, 4094, (char *)cName);
		          strcat_s((char *)cStringa, 4094, "\""); // " is the name string separator
	          }
            COMSendString(COMMAND_SV, (char *)cStringa);
            break;
          case COMMAND_GH: // Get Hover
            lParam = (long)(pg_Param->pVessel->GetThrusterGroupLevel(THGROUP_HOVER) * 100.0 + 0.5);
            COMSendNumber(COMMAND_Ho, lParam);
            pg_Param->lThGroup_HOVER = lParam;
            break;
          case COMMAND_GM: // Get Main
            lParam = (long)(pg_Param->pVessel->GetThrusterGroupLevel(THGROUP_MAIN) * 100.0 + 0.5);
            COMSendNumber(COMMAND_Ma, lParam);
            pg_Param->lThGroup_MAIN = lParam;
            break;
          case COMMAND_GR: // Get Retro
            lParam = (long)(pg_Param->pVessel->GetThrusterGroupLevel(THGROUP_RETRO) * 100.0 + 0.5);
            COMSendNumber(COMMAND_Rt, lParam);
            pg_Param->lThGroup_RETRO = lParam;
            break;
          case COMMAND_GL: // Get RCS Level
            lParam = (long)(pg_Param->rcslevel * 100.0 + 0.5);
            COMSendNumber(COMMAND_Le, lParam);
            break;
          case COMMAND_Gi: // Get KillRot
            sendNewNavmode(NAVMODE_KILLROT,pg_Param->pVessel->GetNavmodeState(NAVMODE_KILLROT));
            break;
          case COMMAND_Ge: // Get HLevel
            sendNewNavmode(NAVMODE_HLEVEL,pg_Param->pVessel->GetNavmodeState(NAVMODE_HLEVEL));
            break;
          case COMMAND_Go: // Get Prograde
            sendNewNavmode(NAVMODE_PROGRADE,pg_Param->pVessel->GetNavmodeState(NAVMODE_PROGRADE));
            break;
          case COMMAND_Gt: // Get Retrograde
            sendNewNavmode(NAVMODE_RETROGRADE,pg_Param->pVessel->GetNavmodeState(NAVMODE_RETROGRADE));
            break;
          case COMMAND_Gp: // Get Normal +
            sendNewNavmode(NAVMODE_NORMAL,pg_Param->pVessel->GetNavmodeState(NAVMODE_NORMAL));
            break;
          case COMMAND_Gm: // Get Normal -
            sendNewNavmode(NAVMODE_ANTINORMAL,pg_Param->pVessel->GetNavmodeState(NAVMODE_ANTINORMAL));
            break;
          case COMMAND_Gh: // Get HAlt
            sendNewNavmode(NAVMODE_HOLDALT,pg_Param->pVessel->GetNavmodeState(NAVMODE_HOLDALT));
            break;
          case COMMAND_CO: // Ready to COM
            pg_Param->pVessel = oapiGetFocusInterface();
            pg_Param->hVessel = pg_Param->pVessel->GetHandle();
            COMSendSimplex(COMMAND_CO);
            break;
          case COMMAND_GT: // Get RCS type
            COMSendNumber(COMMAND_SR,oapiGetAttitudeMode(pg_Param->hVessel));
            break;
          case COMMAND_GF: // Get Focused vessel
            if(pg_Param->pVessel)
            {
              pName = pg_Param->pVessel->GetName();
              COMSendString(COMMAND_Se, pName);
            }
            break;
          }
        }
        lTotReadByte = 0;
        cCommandBuffer[0] = 0;
      }
      else if(lTotReadByte < (COMMANDBUFFERSIZE - 2))
      {
        // uppend cRxBuffer to cCommandBuffer
        cCommandBuffer[lTotReadByte] = cRxBuffer[0];
        lTotReadByte++;
        cCommandBuffer[lTotReadByte] = 0;
      }
      else // cCommandBuffer full, we are getting rubbish
      {
        lTotReadByte = 0;
        cCommandBuffer[0] = 0;
      }
    }
    else//if(nNumberOfBytesRead)
    {
      Sleep(100);
    }
  }//while(true)

#ifdef _DEBUG
      if(pg_Param->hFile && (INVALID_HANDLE_VALUE != pg_Param->hFile))
        CloseHandle(pg_Param->hFile);
      pg_Param->hFile = INVALID_HANDLE_VALUE;
#endif

 pg_Param->bCOMOpened = false;
 pg_Param->bExitCOMThread = false;
}
catch(...)
{
}

_endthread();
}

// ==============================================================
// Local Functions
// ==============================================================

/***********************************/
// Send a command without parameters
void COMSendSimplex(short key)
{
  unsigned char cStringa[16];
  DWORD ulNumberOfBytesWritten;

  if(g_Param.hComm != INVALID_HANDLE_VALUE)
  {
    *(unsigned short *)cStringa = key;
    cStringa[2] = COMMANDTERMINATOR;
    WriteFile(g_Param.hComm, cStringa, 3, &ulNumberOfBytesWritten, NULL);

#ifdef _DEBUG
if(g_Param.hFile)
{
  char cTemp[2];
  cTemp[0] = '[';
  WriteFile(g_Param.hFile,cTemp,1,&ulNumberOfBytesWritten,NULL);
  WriteFile(g_Param.hFile,cStringa,3,&ulNumberOfBytesWritten,NULL);
  cTemp[0] = ']';
  WriteFile(g_Param.hFile,cTemp,1,&ulNumberOfBytesWritten,NULL);
}
#endif
  }
}

/*****************************************/
// Send a command with a numeric parameter
void COMSendNumber(short key, long lNumber)
{
  DWORD ulNumberOfBytesWritten;
  unsigned char cStringa[16];
  unsigned char cStrTemp[16];
  short nLen;

  if(g_Param.hComm != INVALID_HANDLE_VALUE)
  {
    *(unsigned short *)cStringa = key;
    cStringa[2] = 0;
    _itoa_s(lNumber,(char *)cStrTemp,13,10);
    strcat_s((char *)cStringa, 13, (char *)cStrTemp);
    nLen = strlen((char *)cStringa);
    cStringa[nLen] = COMMANDTERMINATOR;
    WriteFile(g_Param.hComm, cStringa, nLen+1, &ulNumberOfBytesWritten, NULL);

#ifdef _DEBUG
if(g_Param.hFile)
{
  char cTemp[2];
  cTemp[0] = '[';
  WriteFile(g_Param.hFile,cTemp,1,&ulNumberOfBytesWritten,NULL);
  WriteFile(g_Param.hFile,cStringa,nLen+1,&ulNumberOfBytesWritten,NULL);
  cTemp[0] = ']';
  WriteFile(g_Param.hFile,cTemp,1,&ulNumberOfBytesWritten,NULL);
}
#endif
  }
}

/****************************************/
// Send a command with a string parameter
void COMSendString(short key, char* pStringa)
{
  DWORD ulNumberOfBytesWritten;
  short nLen;
  unsigned char cStringa[4096];

  if(g_Param.hComm != INVALID_HANDLE_VALUE)
  {
    *(unsigned short *)cStringa = key;
    cStringa[2] = 0;
    strcat_s((char *)cStringa, 4093, (char *)pStringa);
    nLen = strlen((char *)cStringa);
    cStringa[nLen] = COMMANDTERMINATOR;
    WriteFile(g_Param.hComm, cStringa, nLen+1, &ulNumberOfBytesWritten, NULL);

#ifdef _DEBUG
if(g_Param.hFile)
{
  char cTemp[2];
  cTemp[0] = '[';
  WriteFile(g_Param.hFile,cTemp,1,&ulNumberOfBytesWritten,NULL);
  WriteFile(g_Param.hFile,cStringa,nLen+1,&ulNumberOfBytesWritten,NULL);
  cTemp[0] = ']';
  WriteFile(g_Param.hFile,cTemp,1,&ulNumberOfBytesWritten,NULL);
}
#endif
  }
}

// ==============================================================
// Set/reset the navigation mode
void setNewNavmode(int mode, bool active)
{
  if(g_Param.pVessel)
  {
    if(active)
    {
      g_Param.pVessel->ActivateNavmode(mode);
      g_Param.nRCSMode = mode;
    }
    else
    {
      g_Param.pVessel->DeactivateNavmode(mode);
      if(mode == g_Param.nRCSMode)
        g_Param.nRCSMode = 0;
    }
  }
}

// ==============================================================
// Send the nav mode to the palm
void sendNewNavmode(int mode, bool active)
{
  short unCommand;

  if(g_Param.bCOMOpened)
  {
    switch(mode)
    {
    case NAVMODE_KILLROT:
      unCommand = COMMAND_KR;
      break;
    case NAVMODE_HLEVEL:
      unCommand = COMMAND_HL;
      break;
    case NAVMODE_PROGRADE:
      unCommand = COMMAND_Pr;
      break;
    case NAVMODE_RETROGRADE:
      unCommand = COMMAND_Re;
      break;
    case NAVMODE_NORMAL:
      unCommand = COMMAND_NP;
      break;
    case NAVMODE_ANTINORMAL:
      unCommand = COMMAND_NM;
      break;
    case NAVMODE_HOLDALT:
      unCommand = COMMAND_HA;
      break;
    }

    if(active)
    {
      COMSendNumber(unCommand, 1);
      g_Param.nRCSMode = mode;
    }
    else
    {
      COMSendNumber(unCommand, 0);
      if(mode == g_Param.nRCSMode)
        g_Param.nRCSMode = 0;
    }
  }
}

// ==============================================================
// API interfaces
// ==============================================================

// ==============================================================
// Called on focus change
DLLCLBK void opcFocusChanged(OBJHANDLE new_focus, OBJHANDLE old_focus)
{
  char* pName;

  try
  {
    if(g_Param.hVessel)
    {
      g_Param.hVessel = new_focus;
      g_Param.pVessel = oapiGetVesselInterface(g_Param.hVessel);
      pName = g_Param.pVessel->GetName();
      COMSendString(COMMAND_Se, pName);
    }
  }
  catch(...)
  {
  }
}

// ==============================================================
// Called on close
DLLCLBK void opcCloseRenderViewport(void)
{
  try
  {
    g_Param.bExitCOMThread = true;
    Sleep(100);
  }
  catch(...)
  {
  }
}

// ==============================================================
// Executed before the simulation step
DLLCLBK void opcPreStep(double SimT, double SimDT, double mjd)
{
  long  lParam;

  try
  {
    if(g_Param.bCOMOpened)
    {
      if(g_Param.bSustainOn) // To inprove speed response and uniformity on RCS
      {
        // for a uniform RCS control a command is repeated SUSTAINMAXCOUNT steps

        g_Param.nSustainCount++;
        if(g_Param.nSustainCount >= SUSTAINMAXCOUNT)
        {
          g_Param.bSustainOn = false;
          g_Param.nSustainCount = 0;
        }

        if(g_Param.pVessel)
          g_Param.pVessel->IncThrusterGroupLevel_SingleStep((THGROUP_TYPE)g_Param.nSustain, g_Param.rcslevel);
      }

      if((SimT-g_Param.prevSimT) > 0.5) // Every 1/2 second do some check
      {
        g_Param.prevSimT = SimT;

        if(g_Param.pVessel)
        {
          lParam = (long)(g_Param.pVessel->GetThrusterGroupLevel(THGROUP_MAIN) * 100.0 + 0.5);
          if(lParam != g_Param.lThGroup_MAIN)
          {
            g_Param.lThGroup_MAIN = lParam;
            COMSendNumber(COMMAND_Ma, lParam);
          }

          lParam = (long)(g_Param.pVessel->GetThrusterGroupLevel(THGROUP_RETRO) * 100.0 + 0.5);
          if(lParam != g_Param.lThGroup_RETRO)
          {
            g_Param.lThGroup_RETRO = lParam;
            COMSendNumber(COMMAND_Rt, lParam);
          }

          lParam = (long)(g_Param.pVessel->GetThrusterGroupLevel(THGROUP_HOVER) * 100.0 + 0.5);
          if(lParam != g_Param.lThGroup_HOVER)
          {
            g_Param.lThGroup_HOVER = lParam;
            COMSendNumber(COMMAND_Ho, lParam);
          }

          if(g_Param.nRCSMode)
          {
            if(!g_Param.pVessel->GetNavmodeState(g_Param.nRCSMode))
            {
              sendNewNavmode(g_Param.nRCSMode,false);
            }
          }
          else if(g_Param.pVessel->GetNavmodeState(NAVMODE_KILLROT))
          {
            sendNewNavmode(NAVMODE_KILLROT,true);
          }
          else if(g_Param.pVessel->GetNavmodeState(NAVMODE_HLEVEL))
          {
            sendNewNavmode(NAVMODE_HLEVEL,true);
          }
          else if(g_Param.pVessel->GetNavmodeState(NAVMODE_PROGRADE))
          {
            sendNewNavmode(NAVMODE_PROGRADE,true);
          }
          else if(g_Param.pVessel->GetNavmodeState(NAVMODE_RETROGRADE))
          {
            sendNewNavmode(NAVMODE_RETROGRADE,true);
          }
          else if(g_Param.pVessel->GetNavmodeState(NAVMODE_NORMAL))
          {
            sendNewNavmode(NAVMODE_NORMAL,true);
          }
          else if(g_Param.pVessel->GetNavmodeState(NAVMODE_ANTINORMAL))
          {
            sendNewNavmode(NAVMODE_ANTINORMAL,true);
          }
          else if(g_Param.pVessel->GetNavmodeState(NAVMODE_HOLDALT))
          {
            sendNewNavmode(NAVMODE_HOLDALT,true);
          }
        }
      }
    }
  }
  catch(...)
  {
  }
}

// ==============================================================
// This function is called when Orbiter starts or when the module
// is activated.
DLLCLBK void InitModule(HINSTANCE hDLL)
{
  try
  {
    g_Param.g_hInst = hDLL; // remember the instance handle

    // To allow the user to open our new dialog box, we create
    // an entry in the "Custom Functions" list which is accessed
    // in Orbiter via Ctrl-F4.
    g_Param.g_dwCmd = oapiRegisterCustomCmd("Palm Remote Control",
      "Control a vessel with a Palm through a serial connection",
      OpenDlgClbk, NULL);

    g_Param.nRCSMode = 0;
    g_Param.pVessel = NULL;
    g_Param.hVessel = NULL;
    g_Param.bCOMOpened = false;
    g_Param.bExitCOMThread = false;
    g_Param.bSustainOn = false;
    g_Param.nSustainCount = 0;
    g_Param.nSustain = 0;
    g_Param.nRCSMode = 0;
    g_Param.iComNum = 3;
    g_Param.rcslevel = 1.0;
    g_Param.prevSimT = 0;
    g_Param.lThGroup_MAIN = 0;
    g_Param.lThGroup_RETRO = 0;
    g_Param.lThGroup_HOVER = 0;
    g_Param.hComm = INVALID_HANDLE_VALUE;
#ifdef _DEBUG
    g_Param.hFile = INVALID_HANDLE_VALUE;
#endif
  }
  catch(...)
  {
  }
}

// ==============================================================
// This function is called when Orbiter shuts down or when the
// module is deactivated
DLLCLBK void ExitModule(HINSTANCE hDLL)
{
  try
  {
    g_Param.bExitCOMThread = true;
    Sleep(300);

    // Unregister the custom function in Orbiter
    oapiUnregisterCustomCmd(g_Param.g_dwCmd);
  }
  catch(...)
  {
  }
}

// ==============================================================
// Write some parameters to the scenario file
DLLCLBK void opcSaveState(FILEHANDLE scn)
{
  try
  {
    oapiWriteScenario_int(scn, "ParComNum", g_Param.iComNum);
  }
  catch(...)
  {
  }
}

// ==============================================================
// Read custom parameters from scenario
DLLCLBK void opcLoadState(FILEHANDLE scn)
{
  char *line;

  try
  {
    while(oapiReadScenario_nextline(scn, line))
    {
      if(!_strnicmp(line, "ParComNum", 5))
      {
        sscanf_s(line+9, "%d", &g_Param.iComNum);
      }
    }
  }
  catch(...)
  {
  }
}

// ==============================================================
// Open the dialog window
void OpenDlgClbk(void *context)
{
  try
  {
    HWND hDlg = oapiOpenDialog(g_Param.g_hInst, IDD_PALMRCONTROL, MsgProc);
    // Don't use a standard Windows function like CreateWindow to
    // open the dialog box, because it won't work in fullscreen mode
  }
  catch(...)
  {
  }
}

// ==============================================================
// Close the dialog
void CloseDlg(HWND hDlg)
{
  try
  {
    // do not close COM if open
    oapiCloseDialog(hDlg);
  }
  catch(...)
  {
  }
}

// ==============================================================
// Windows message handler for the dialog box
BOOL CALLBACK MsgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  char cStringa[256];

  try
  {
    switch(uMsg)
    {
    case WM_INITDIALOG:
      sprintf_s(cStringa, "%d", g_Param.iComNum);
      SetWindowText(GetDlgItem(hDlg, IDC_SERNUM), cStringa);
      if(g_Param.bCOMOpened)
        sprintf_s(cStringa, "%s", "Open");
      else
        sprintf_s(cStringa, "%s", "Closed");
      SetWindowText(GetDlgItem(hDlg, IDC_STATE), cStringa);

      g_Param.pVessel = oapiGetFocusInterface();
      g_Param.hVessel = g_Param.pVessel->GetHandle();
      return TRUE;

    case WM_DESTROY:
      GetWindowText(GetDlgItem(hDlg, IDC_SERNUM), cStringa, 256);
      sscanf_s(cStringa, "%d", &g_Param.iComNum);
      return TRUE;

    case WM_COMMAND:
      switch(LOWORD(wParam))
      {
      case IDC_OPENSER: // open serial connection
        if(g_Param.bCOMOpened) // previous thread still running
          return TRUE;
        GetWindowText(GetDlgItem(hDlg, IDC_SERNUM), cStringa, 256);
        sscanf_s(cStringa, "%d", &g_Param.iComNum);
        sprintf_s(cStringa, "COM%d", g_Param.iComNum);
        // open serial port
        g_Param.hComm = CreateFile( cStringa,  
                                    GENERIC_READ | GENERIC_WRITE, 
                                    0, 
                                    0, 
                                    OPEN_EXISTING,
                                    0,
                                    0);
        if(g_Param.hComm == INVALID_HANDLE_VALUE)
        {
          sprintf_s(cStringa, "%s", "Open COM Error");
        }
        else
        {
          sprintf_s(cStringa, "%s", "Open");
          // get the default COM timeout and set the new ones
          GetCommTimeouts(g_Param.hComm,&g_Param.CommDefaultTimeouts);
          g_Param.CommTimeouts.ReadIntervalTimeout = 20;
          g_Param.CommTimeouts.ReadTotalTimeoutMultiplier = 10;
          g_Param.CommTimeouts.ReadTotalTimeoutConstant = 100;
          g_Param.CommTimeouts.WriteTotalTimeoutMultiplier = 10;
          g_Param.CommTimeouts.WriteTotalTimeoutConstant = 100;
          if(!SetCommTimeouts(g_Param.hComm, &g_Param.CommTimeouts))
          {
            sprintf_s(cStringa, "%s", "Set COM Error");
            CloseHandle(g_Param.hComm);// close serial COM
            g_Param.hComm = INVALID_HANDLE_VALUE;
            g_Param.bCOMOpened = false;
          }
          else
          {
            PurgeComm(g_Param.hComm, PURGE_TXCLEAR);
            PurgeComm(g_Param.hComm, PURGE_RXCLEAR);
            g_Param.bCOMOpened = true;
            g_Param.bExitCOMThread = false;
            // create COM thread
            _beginthread( COMThread, 0, &g_Param );
          }
        }
        SetWindowText(GetDlgItem(hDlg, IDC_STATE), cStringa);
        return TRUE;

      case IDC_CLOSESER: // close serial connection
        if(!g_Param.bCOMOpened) // already closed
          return TRUE;
        cStringa[0] = 0;			
        SetWindowText(GetDlgItem(hDlg, IDC_DATAIN), cStringa);
        sprintf_s(cStringa, "%s", "Closed");
        SetWindowText(GetDlgItem(hDlg, IDC_STATE), cStringa);
        // close COM
        g_Param.bExitCOMThread = true;
        return TRUE;

      case IDCANCEL: // dialog closed by user
        CloseDlg(hDlg);
        return TRUE;
      }
      break;
    }
    return oapiDefDialogProc(hDlg, uMsg, wParam, lParam);
  }
  catch(...)
  {
  }
  return false;
}
